 /*----------------------------------------------------------------------------------------------------+
 |                             National Institute of Standards and Technology                          |
 |                                        Biometric Clients Lab                                        |
 +-----------------------------------------------------------------------------------------------------+
  File author(s):
       Kevin Mangold (kevin.mangold@nist.gov)
 
 +-----------------------------------------------------------------------------------------------------+
 | NOTICE & DISCLAIMER                                                                                 |
 |                                                                                                     |
 | The research software provided on this web site ("software") is provided by NIST as a public        |
 | service. You may use, copy and distribute copies of the software in any medium, provided that you   |
 | keep intact this entire notice. You may improve, modify and create derivative works of the software |
 | or any portion of the software, and you may copy and distribute such modifications or works.        |
 | Modified works should carry a notice stating that you changed the software and should note the date |
 | and nature of any such change.  Please explicitly acknowledge the National Institute of Standards   |
 | and Technology as the source of the software.                                                       |
 |                                                                                                     |
 | The software is expressly provided "AS IS."  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED,  |
 | IN FACT OR ARISING BY OPERATION OF LAW, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF      |
 | MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.  NIST        |
 | NEITHER REPRESENTS NOR WARRANTS THAT THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR         |
 | ERROR-FREE, OR THAT ANY DEFECTS WILL BE CORRECTED.  NIST DOES NOT WARRANT OR MAKE ANY               |
 | REPRESENTATIONS REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED |
 | TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.                           |
 |                                                                                                     |
 | You are solely responsible for determining the appropriateness of using and distributing the        |
 | software and you assume all risks associated with its use, including but not limited to the risks   |
 | and costs of program errors, compliance with applicable laws, damage to or loss of data, programs   |
 | or equipment, and the unavailability or interruption of operation.  This software is not intended   |
 | to be used in any situation where a failure could cause risk of injury or damage to property.  The  |
 | software was developed by NIST employees.  NIST employee contributions are not subject to copyright |
 | protection within the United States.                                                                |
 |                                                                                                     |
 | Specific hardware and software products identified in this open source project were used in order   |
 | to perform technology transfer and collaboration. In no case does such identification imply         |
 | recommendation or endorsement by the National Institute of Standards and Technology, nor            |
 | does it imply that the products and equipment identified are necessarily the best available for the |
 | purpose.                                                                                            |
 +----------------------------------------------------------------------------------------------------*/

package gov.nist.itl.wsbd.test.tests;

import static gov.nist.itl.wsbd.Logging.trace;
import gov.nist.itl.wsbd.dom.Dictionary;
import gov.nist.itl.wsbd.persistence.CountCapacityExceededException;
import gov.nist.itl.wsbd.persistence.FileStorageProvider;
import gov.nist.itl.wsbd.persistence.IStorageProvider;
import gov.nist.itl.wsbd.persistence.PersistenceState;
import gov.nist.itl.wsbd.persistence.StorageCapacityExceededException;

import java.io.File;
import java.util.UUID;

import junit.framework.Assert;

import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;

public class FileStorageProviderTests {
	public static IStorageProvider provider;
	public static final String location = System.getProperty("java.io.tmpdir") + "/WSBD";
	public static final long dataCapacity = Long.MAX_VALUE;
	public static final int countCapacity = Integer.MAX_VALUE;
	
	@BeforeClass
	public static void setUp() {
		resetStorageLocation();
	}
	
	protected static void resetStorageLocation() {
		resetStorageLocation(Long.MAX_VALUE, Integer.MAX_VALUE, false);
	}
	
	protected static void resetStorageLocation(long dataCapacity, int countCapacity, boolean lru) {
		provider = new FileStorageProvider(location, dataCapacity, countCapacity, lru);
		new File(location).mkdirs();
	}
	
	@AfterClass
	public static void tearDown() {
		File f = new File(location);
		clean(f);
	}
	
	protected static void clean(String path) {
		clean(new File(path));
	}
	
	protected static void clean(File f) {
		if (f.isDirectory()) {
			for (File c : f.listFiles()) {
				clean(c);
			}
		}
		
		f.delete();
	}
	
	@Test
	public void ReservationOfUnusedIdIsSuccessful() {
		trace("FileStorageProviderTests.ReservationOfUnusedIdIsSuccessful()");
		
		resetStorageLocation();
		
		UUID id = UUID.randomUUID();
		
		Assert.assertEquals(PersistenceState.EMPTY, provider.getState(id));
		Assert.assertTrue(provider.reserveId(id));
		Assert.assertEquals(PersistenceState.PENDING, provider.getState(id));
		Assert.assertFalse(provider.reserveId(id));
	}
	
	@Test
	public void StoredDataHasCorrectState() {
		trace("FileStorageProviderTests.StoredDataHasCorrectState()");
		
		resetStorageLocation();
		
		byte[] b = new byte[] {  };
		Dictionary d = new Dictionary();
		
		UUID id = UUID.randomUUID();

		while(!provider.reserveId(id)) {
			id = UUID.randomUUID();
		}
		
		Assert.assertEquals(PersistenceState.PENDING, provider.getState(id));
		provider.store(id, d, b);
		Assert.assertEquals(PersistenceState.OK, provider.getState(id));
	}
	
	@Test (expected = CountCapacityExceededException.class)
	public void CountLimitIsEnforced() {
		trace("FileStorageProviderTests.CountLimitIsEnforced()");
		
		resetStorageLocation();
		
		int testCount = 2;
		
		FileStorageProvider localProvider = new FileStorageProvider(location, dataCapacity, testCount);
		
		byte[] b = new byte[] {  };
		Dictionary d = new Dictionary();
		
		for (int i = 0; i < testCount; i++) {
			UUID id = UUID.randomUUID();
			localProvider.store(id, d, b);
		}
	}
	
	@Test (expected = StorageCapacityExceededException.class)
	public void SizeLimitIsEnforced() {
		trace("FileStorageProviderTests.SizeLimitIsEnforced()");
		
		resetStorageLocation();
		
		FileStorageProvider localProvider = new FileStorageProvider(location, 1, countCapacity);
		
		byte[] b = new byte[] {  };
		Dictionary d = new Dictionary();		
		UUID id = UUID.randomUUID();
		localProvider.store(id, d, b);
	}
	
	@Test
	public void LRUDataIsDroppedFromStorageCapacity() {
		trace("FileStorageProviderTests.LRUDataIsDroppedFromStorageCapacity()");
		
		resetStorageLocation(10240, Integer.MAX_VALUE, true);
		
		byte[] b = new byte[] {};
		Dictionary d = new Dictionary();
		
		for (int i = 0; i < 10; i++) {
			UUID id = UUID.randomUUID();
			provider.store(id, d, b);
		}
	}
	
	@Test
	public void LRUDataIsDroppedFromCountCapacity() {
		trace("FileStorageProviderTests.LRUDataIsDroppedFromCountCapacity()");
		
		int testCount = 6;
		resetStorageLocation(Long.MAX_VALUE, testCount, true);
		
		byte[] b = new byte[] {1, 2, 3, 4, 5};
		Dictionary d = new Dictionary();
		
		for (int i = 0; i < testCount; i++) {
			UUID id = UUID.randomUUID();
			provider.store(id, d, b);
		}
	}
}