 /*----------------------------------------------------------------------------------------------------+
 |                             National Institute of Standards and Technology                          |
 |                                        Biometric Clients Lab                                        |
 +-----------------------------------------------------------------------------------------------------+
  File author(s):
       Kevin Mangold (kevin.mangold@nist.gov)
 
 +-----------------------------------------------------------------------------------------------------+
 | NOTICE & DISCLAIMER                                                                                 |
 |                                                                                                     |
 | The research software provided on this web site ("software") is provided by NIST as a public        |
 | service. You may use, copy and distribute copies of the software in any medium, provided that you   |
 | keep intact this entire notice. You may improve, modify and create derivative works of the software |
 | or any portion of the software, and you may copy and distribute such modifications or works.        |
 | Modified works should carry a notice stating that you changed the software and should note the date |
 | and nature of any such change.  Please explicitly acknowledge the National Institute of Standards   |
 | and Technology as the source of the software.                                                       |
 |                                                                                                     |
 | The software is expressly provided "AS IS."  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED,  |
 | IN FACT OR ARISING BY OPERATION OF LAW, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF      |
 | MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.  NIST        |
 | NEITHER REPRESENTS NOR WARRANTS THAT THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR         |
 | ERROR-FREE, OR THAT ANY DEFECTS WILL BE CORRECTED.  NIST DOES NOT WARRANT OR MAKE ANY               |
 | REPRESENTATIONS REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED |
 | TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.                           |
 |                                                                                                     |
 | You are solely responsible for determining the appropriateness of using and distributing the        |
 | software and you assume all risks associated with its use, including but not limited to the risks   |
 | and costs of program errors, compliance with applicable laws, damage to or loss of data, programs   |
 | or equipment, and the unavailability or interruption of operation.  This software is not intended   |
 | to be used in any situation where a failure could cause risk of injury or damage to property.  The  |
 | software was developed by NIST employees.  NIST employee contributions are not subject to copyright |
 | protection within the United States.                                                                |
 |                                                                                                     |
 | Specific hardware and software products identified in this open source project were used in order   |
 | to perform technology transfer and collaboration. In no case does such identification imply         |
 | recommendation or endorsement by the National Institute of Standards and Technology, nor            |
 | does it imply that the products and equipment identified are necessarily the best available for the |
 | purpose.                                                                                            |
 +----------------------------------------------------------------------------------------------------*/

package gov.nist.itl.wsbd.test;

import static gov.nist.itl.wsbd.Logging.trace;
import gov.nist.itl.wsbd.SensorService;
import gov.nist.itl.wsbd.Utility;
import gov.nist.itl.wsbd.concurrency.SessionManager;
import gov.nist.itl.wsbd.dom.Dictionary;
import gov.nist.itl.wsbd.dom.Result;
import gov.nist.itl.wsbd.dom.Status;
import gov.nist.itl.wsbd.persistence.BindingContext;
import gov.nist.itl.wsbd.persistence.FileStorageProvider;
import gov.nist.itl.wsbd.persistence.IStorageProvider;
import gov.nist.itl.wsbd.persistence.PersistenceState;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URISyntaxException;
import java.util.UUID;

import javax.ws.rs.Path;
import javax.ws.rs.core.Response;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;

@Path("/")
public class TestService extends SensorService {
	protected static Dictionary serviceInformation;
	protected static Dictionary configuration;
	protected static boolean initialized = false;
	protected static IStorageProvider storage;
	
	public TestService() {
		trace("TestService()");
		
		if (storage == null) {
			storage = new FileStorageProvider(System.getProperty("java.io.tmpdir") + "/WSBD_TestService", Long.MAX_VALUE, Integer.MAX_VALUE);
		}
		
		if (serviceInformation == null) {
			serviceInformation = new Dictionary();
			initializeServiceInformation();
		}
		
		if (configuration == null) {
			configuration = new Dictionary();
			configuration.put("ImageWidth", "640");
			configuration.put("ImageHeight", "480");
			configuration.put("Vendor", "National Institute of Standards and Technology");
		}
	}
	
	@Override
	public Result register() {
		trace("TestService.register()");
		
		initialized = false;
		return super.register();
	}
	
	protected static boolean isInitialized() {
		trace("TestService.isInitialized()");
		
		return initialized;
	}
	
	protected static void initializeServiceInformation() {
		trace("TestService.initializeServiceInformation()");
		
		try {
			Unmarshaller u = BindingContext.getBindingContext(Dictionary.class).createUnmarshaller();
			InputStream is = TestService.class.getResourceAsStream(TestConstants.ServiceInformationFilePath);
			Object obj = u.unmarshal(is);
			serviceInformation = (Dictionary)obj;
		} catch (JAXBException jaxbe) {
			jaxbe.printStackTrace();
		}
	}
	
	@Override
	public Result getServiceInformation() {
		trace("TestService.getServiceInformation()");
		
		Result result = new Result(Status.SUCCESS);
		result.setMetadata(serviceInformation);
		return result;
	}

	@Override
	public Result initialize(UUID sessionId) throws InterruptedException {
		trace("TestService.initialize(UUID)");
		
		Result result = new Result();
		
		if (SessionManager.getInstance(this).isRegistered(sessionId)) {
			if (getServiceLock(this).isLocked()) {
				if (getServiceLock(this).holdsLock(sessionId)) {
					//try {
						Thread.sleep(TestConstants.initializationDuration);
						result.setStatus(Status.SUCCESS);
						initialized = true;
					//} catch (InterruptedException ie) {
						//ie.printStackTrace();
						//result.setStatus(Status.CANCELED);
					//}
				} else {
					result.setStatus(Status.LOCK_HELD_BY_ANOTHER);
				}
			} else {
				result.setStatus(Status.LOCK_NOT_HELD);
			}
		} else {
			result.setStatus(Status.FAILURE);
		}

		return result;
	}

	@Override
	public Result getConfiguration(UUID sessionId) throws InterruptedException {
		trace("TestService.getConfiguration(UUID)");
		
		Result result = new Result();
		
		if (SessionManager.getInstance(this).isRegistered(sessionId)) {
			if (getServiceLock(this).isLocked()) {
				if (getServiceLock(this).holdsLock(sessionId)) {
					if (isInitialized()) {
						result.setStatus(Status.SUCCESS);
						result.setMetadata(configuration);
						
						Thread.sleep(TestConstants.getConfigurationDuration);
					} else {
						result.setStatus(Status.INITIALIZATION_NEEDED);
					}
				} else {
					result.setStatus(Status.LOCK_HELD_BY_ANOTHER);
				}
			} else {
				result.setStatus(Status.LOCK_NOT_HELD);
			}
		} else {
			if (sessionId.equals(UUID.fromString(TestConstants.invalidSessionId))) {
				result.setStatus(Status.INVALID_ID);
				result.setBadFields(new String[] { "sessionId" });
			} else {
				result.setStatus(Status.FAILURE);
			}
		}
		
		return result;
	}

	@Override
	public Result setConfiguration(UUID sessionId, Dictionary configuration) throws InterruptedException {
		trace("TestService.setConfiguration(UUID, Dictionary)");
		
		Result result = new Result();
		
		if (SessionManager.getInstance(this).isRegistered(sessionId)) {
			if (getServiceLock(this).isLocked()) {
				if (getServiceLock(this).holdsLock(sessionId)) {
					if (isInitialized()) {
						for (String key : configuration.keySet()) {
							TestService.configuration.put(key, configuration.get(key));
						}
						
						result.setStatus(Status.SUCCESS);
						
						Thread.sleep(TestConstants.setConfigurationDuration);
					} else {
						result.setStatus(Status.INITIALIZATION_NEEDED);
					}
				} else {
					result.setStatus(Status.LOCK_HELD_BY_ANOTHER);
				}
			} else {
				result.setStatus(Status.LOCK_NOT_HELD);
			}
		} else {
			if (sessionId.equals(UUID.fromString(TestConstants.invalidSessionId))) {
				result.setStatus(Status.INVALID_ID);
				result.setBadFields(new String[] { "sessionId" });
			} else {
				result.setStatus(Status.FAILURE);
			}
		}
		
		return result;
	}

	@Override
	public Result capture(UUID sessionId) throws InterruptedException {
		trace("TestService.capture(UUID)");
		
		UUID tempCaptureId;
		
		do {
			tempCaptureId = UUID.randomUUID();
		} while (!storage.reserveId(tempCaptureId));
		
		final UUID captureId = tempCaptureId;
		
		do {
			tempCaptureId = UUID.randomUUID();
		} while (!storage.reserveId(tempCaptureId));
		
		final UUID captureId2 = tempCaptureId;
		
		Result result = new Result();

		if (SessionManager.getInstance(this).isRegistered(sessionId)) {
			if (getServiceLock(this).isLocked()) {
				if (getServiceLock(this).holdsLock(sessionId)) {
					if (isInitialized()) {
						try {
							byte[] buffer = Utility.readAllAsBytes(TestService.class.getResourceAsStream(TestConstants.SampleFingerPrintFilePath));

							Dictionary metadata = new Dictionary();
							metadata.put("captureTime", TestConstants.timestamp());
							metadata.put("modality", "Finger");
							metadata.put("submodality", "flatLeftIndex");
							metadata.put("contentType", "image/jpeg");

							Thread.sleep(TestConstants.captureDuration);

							storage.store(captureId, metadata, buffer);
							storage.store(captureId2, metadata, buffer);
						} catch (IOException ioe) {
							ioe.printStackTrace();
						}
						
						result.setCaptureIds(new UUID[] { captureId, captureId2 });
						result.setStatus(Status.SUCCESS);
					} else {
						result.setStatus(Status.INITIALIZATION_NEEDED);
					}
				} else {
					result.setStatus(Status.LOCK_HELD_BY_ANOTHER);
				}
			} else {
				result.setStatus(Status.LOCK_NOT_HELD);
			}
		} else {
			if (sessionId.equals(UUID.fromString(TestConstants.invalidSessionId))) {
				result.setStatus(Status.INVALID_ID);
				result.setBadFields(new String[] { "sessionId" });
			} else {
				result.setStatus(Status.FAILURE);
			}
		}
		
		// TODO TestService.capture()
		//canceled
		//canceledWithSensorFailure
		//sensorFailure
		//configurationNeeded
		//sensorBusy
		//sensorTimeout
		//badValue
		return result;
	}

	@Override
	public Result download(UUID captureId) {
		trace("TestService.download(UUID)");
		
		Result result = new Result();
		
		if (captureId.equals(UUID.fromString(TestConstants.invalidSessionId))) {
			result.setStatus(Status.INVALID_ID);
		} else {
			PersistenceState status = storage.getState(captureId);
			
			switch (status) {
				case EMPTY:
					if (Utility.isValidCaptureId(captureId)) {
						result.setStatus(Status.INVALID_ID);
					} else {
						result.setStatus(Status.BAD_VALUE);
					}
					
					result.setBadFields(new String[] { "captureId" });
					break;
				case PENDING:
					result.setStatus(Status.PREPARING_DOWNLOAD);
					break;
				case CORRUPT:
					result.setStatus(Status.FAILURE);
					break;
				case OK:
					result.setStatus(Status.SUCCESS);
					result.setMetadata(storage.getMetadata(captureId));
					result.setSensorData(storage.getData(captureId));
					break;
				default:
			}
		}
		
		return result;
	}

	@Override
	public Result getDownloadInformation(UUID captureId) {
		trace("TestService.getDownloadInformation(UUID)");
		
		Result result = new Result();
		
		if (captureId.equals(UUID.fromString(TestConstants.invalidSessionId))) {
			result.setStatus(Status.INVALID_ID);
		} else {
			PersistenceState status = storage.getState(captureId);
			
			switch (status) {
				case EMPTY:
					result.setStatus(Status.BAD_VALUE);
					result.setBadFields(new String[] { "captureId" });
					break;
				case PENDING:
					result.setStatus(Status.PREPARING_DOWNLOAD);
					break;
				case CORRUPT:
					result.setStatus(Status.FAILURE);
					break;
				case OK:
					result.setStatus(Status.SUCCESS);
					result.setMetadata(storage.getMetadata(captureId));
					break;
				default:
			}
		}
		
		return result;
	}

	@Override
	public Result thriftyDownload(UUID captureId, String maxSize) {
		trace("TestService.thriftyDownload(UUID, String)");
		
		// TODO TestService.thriftyDownload()
		//success
		//failure
		//invalidId
		//badValue
		//preparingDownload
		return download(captureId);
	}
	

	@Override
	public InputStream stream() {
		return null;
	}
}