 /*----------------------------------------------------------------------------------------------------+
 |                             National Institute of Standards and Technology                          |
 |                                        Biometric Clients Lab                                        |
 +-----------------------------------------------------------------------------------------------------+
  File author(s):
       Kevin Mangold (kevin.mangold@nist.gov)
 
 +-----------------------------------------------------------------------------------------------------+
 | NOTICE & DISCLAIMER                                                                                 |
 |                                                                                                     |
 | The research software provided on this web site ("software") is provided by NIST as a public        |
 | service. You may use, copy and distribute copies of the software in any medium, provided that you   |
 | keep intact this entire notice. You may improve, modify and create derivative works of the software |
 | or any portion of the software, and you may copy and distribute such modifications or works.        |
 | Modified works should carry a notice stating that you changed the software and should note the date |
 | and nature of any such change.  Please explicitly acknowledge the National Institute of Standards   |
 | and Technology as the source of the software.                                                       |
 |                                                                                                     |
 | The software is expressly provided "AS IS."  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED,  |
 | IN FACT OR ARISING BY OPERATION OF LAW, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF      |
 | MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.  NIST        |
 | NEITHER REPRESENTS NOR WARRANTS THAT THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR         |
 | ERROR-FREE, OR THAT ANY DEFECTS WILL BE CORRECTED.  NIST DOES NOT WARRANT OR MAKE ANY               |
 | REPRESENTATIONS REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED |
 | TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.                           |
 |                                                                                                     |
 | You are solely responsible for determining the appropriateness of using and distributing the        |
 | software and you assume all risks associated with its use, including but not limited to the risks   |
 | and costs of program errors, compliance with applicable laws, damage to or loss of data, programs   |
 | or equipment, and the unavailability or interruption of operation.  This software is not intended   |
 | to be used in any situation where a failure could cause risk of injury or damage to property.  The  |
 | software was developed by NIST employees.  NIST employee contributions are not subject to copyright |
 | protection within the United States.                                                                |
 |                                                                                                     |
 | Specific hardware and software products identified in this open source project were used in order   |
 | to perform technology transfer and collaboration. In no case does such identification imply         |
 | recommendation or endorsement by the National Institute of Standards and Technology, nor            |
 | does it imply that the products and equipment identified are necessarily the best available for the |
 | purpose.                                                                                            |
 +----------------------------------------------------------------------------------------------------*/

package gov.nist.itl.wsbd.persistence;

import static gov.nist.itl.wsbd.Logging.trace;
import gov.nist.itl.wsbd.dom.Dictionary;

import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;

public class FileStorageProvider implements IStorageProvider {
	protected final String location;
	protected final long dataCapacity;
	protected final int countCapacity;
	protected final boolean leastUsedDataDropped;
	protected volatile long calculatedSize;
	protected Map<UUID, Long> activityRecord;
	
	public FileStorageProvider() {
		this(System.getProperty("java.io.tmpdir") + "/WSBD", Long.MAX_VALUE, Integer.MAX_VALUE);
	}
	
	public FileStorageProvider(String location, long dataCapacity, int countCapacity) {
		this(location, dataCapacity, countCapacity, false);
	}
	
	public FileStorageProvider(String location, long dataCapacity, int countCapacity, boolean leastUsedDataDropped) {
		trace("FileStorageProvider(String, long, int, boolean)");
		
		this.location = location;
		this.dataCapacity = dataCapacity;
		this.countCapacity = countCapacity;
		this.leastUsedDataDropped = leastUsedDataDropped;

		activityRecord = new HashMap<UUID, Long>();
		
		File f = new File(location);
		clean(f);
		f.mkdirs();
		
		validateCountCapacity();
		validateStorageCapacity();
	}
	
	protected static void clean(String path) {
		trace("FileStorageProvider.clean(String)");
		
		clean(new File(path));
	}
	
	protected static void clean(File f) {
		trace("FileStorageProvider.clean(File)");
		
		if (f.isDirectory()) {
			for (File c : f.listFiles()) {
				clean(c);
			}
		}
		
		f.delete();
	}
	
	protected String getFilePath(UUID id, String extension) {
		trace("FileStorageProvider.getFilePath(UUID, String)");
		
		return String.format("%s/%s.%s", location, id.toString(), extension);
	}
	
	protected String getDataFilePath(UUID id) {
		return getFilePath(id, "data");
	}
	
	protected String getMetadataFilePath(UUID id) {	
		return getFilePath(id, "info");
	}
	
	protected String getReservationFilePath(UUID id) {
		return getFilePath(id, "tmp");
	}
	
	protected boolean exists(String filepath) {
		return new File(filepath).exists();
	}
	
	protected void update(UUID id) {
		trace("FileStorageProvider.update(UUID)");
		
		synchronized (activityRecord) {
			activityRecord.put(id, System.currentTimeMillis());
		}
	}

	@Override
	public boolean reserveId(UUID id) {
		trace("FileStorageProvider.reserveId(UUID)");
		
		if (getState(id) != PersistenceState.EMPTY) {
			return false;
		} else {
			File f = new File(getReservationFilePath(id));
			
			try {
				f.createNewFile();
				return true;
			} catch (IOException ioe) {
				ioe.printStackTrace();
				return false;
			}
		}
	}

	@Override
	public PersistenceState getState(UUID id) {
		trace("FileStorageProvider.getState(UUID)");
		
		if (exists(getReservationFilePath(id))) {
			return PersistenceState.PENDING;
		} else if (exists(getDataFilePath(id)) && exists(getMetadataFilePath(id))) {
			return PersistenceState.OK;
		} else if (exists(getDataFilePath(id)) ^ exists(getMetadataFilePath(id))) {
			return PersistenceState.CORRUPT;
		} else {
			return PersistenceState.EMPTY;
		}
	}

	@Override
	public void store(UUID id, Dictionary metadata, byte[] data) {
		trace("FileStorageProvider.store(UUID, Dictionary, byte[])");
		
		update(id);
		
		validateCountCapacity();
		validateStorageCapacity();
		
		PersistenceState state = getState(id);
		
		if (state == PersistenceState.PENDING || state == PersistenceState.EMPTY) {
			try {
				validateStorageCapacity(data.length);

				DataOutputStream out = new DataOutputStream(new FileOutputStream(new File(getDataFilePath(id))));
				out.write(data);
				out.close();
				
				calculatedSize += data.length;
			} catch (FileNotFoundException fnfe) {
				fnfe.printStackTrace();
			} catch (IOException ioe) {
				ioe.printStackTrace();
			}

			try {
				JAXBContext context = BindingContext.getBindingContext(Dictionary.class);
				
				long size = getSizeInBytes(metadata);
				validateStorageCapacity(size);
				
				Marshaller m = context.createMarshaller();
				m.marshal(metadata, new File(getMetadataFilePath(id)));
				
				calculatedSize += size;
			} catch (JAXBException jaxbe) {
				jaxbe.printStackTrace();
			}
		}
		
		File f = new File(getReservationFilePath(id));
		while (!f.delete()) {
			if (!f.exists())
				break;
			
			System.err.printf("Cannot delete reservation file: %s%n", f.getPath());
		}
	}

	@Override
	public byte[] getData(UUID id) {
		trace("FileStorageProvider.getData(UUID)");
		
		update(id);
		
		File f = new File(getDataFilePath(id));
		long fileLength = f.length();
		
		byte[] buffer = null;
		
		if (fileLength <= Integer.MAX_VALUE) {
			buffer = new byte[(int)fileLength];
		} else {
			return null;
		}

		try {
			DataInputStream dis = new DataInputStream(new FileInputStream(f));
			dis.readFully(buffer);
			dis.close();
		} catch (FileNotFoundException fnfe) {
			fnfe.printStackTrace();
			return null;
		} catch (IOException ioe) {
			ioe.printStackTrace();
			return null;
		}
		
		return buffer;
	}

	@Override
	public Dictionary getMetadata(UUID id) {
		trace("FileStorageProvider.getMetadata(UUID)");
		
		update(id);
		
		Dictionary d = null;
		
		try {
			JAXBContext context = BindingContext.getBindingContext(Dictionary.class);
			Unmarshaller u = context.createUnmarshaller();
			d = (Dictionary)u.unmarshal(new File(getMetadataFilePath(id)));
		} catch (JAXBException jaxbe) {
			jaxbe.printStackTrace();
		}
		
		return d;
	}
	
	protected void validateStorageCapacity() {
		trace("FileStorageProvider.validateStorageCapacity()");
		
		validateStorageCapacity(false);
	}
	
	protected void validateStorageCapacity(long additionalBytes) {
		trace("FileStorageProvider.validateStorageCapacity(long)");
		
		if (calculatedSize + additionalBytes >= dataCapacity) {
			if (leastUsedDataDropped) {
				deleteLeastRecentlyUsed();
				validateStorageCapacity(additionalBytes);
			} else {
				throw new StorageCapacityExceededException();
			}
		}
	}
	
	protected void validateStorageCapacity(boolean recalculate) {
		trace("FileStorageProvider.validateStorageCapacity(boolean)");
		
		if (recalculate) {
			File root = new File(location);
			calculatedSize = 0;
			
			if (root.exists()) {
				for (File f : root.listFiles()) {
					calculatedSize += f.length();
				}
			}
		}
		
		if (calculatedSize >= dataCapacity) {
			if (leastUsedDataDropped) {
				deleteLeastRecentlyUsed();
				validateStorageCapacity();
			} else {
				throw new StorageCapacityExceededException();
			}
		}
	}
	
	protected void validateCountCapacity() {
		trace("FileStorageProvider.validateCountCapacity()");
		
		File f = new File(location);
		
		if (f.exists()) {
			f.listFiles();
			long l = f.listFiles().length;
			if (f.listFiles().length >= countCapacity) {
				if (leastUsedDataDropped) {
					deleteLeastRecentlyUsed();
					validateCountCapacity();
				} else {
					throw new CountCapacityExceededException();
				}
			}
		}
	}
	
	protected long getSizeInBytes(Object obj) {
		trace("FileStorageProvider.getSizeInBytes(Object)");
		
		long size = 0;
		
		try {
			JAXBContext context = BindingContext.getBindingContext(obj.getClass());
			Marshaller m = context.createMarshaller();
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			m.marshal(obj, baos);
			String s = new String(baos.toByteArray());
			size = s.length();
		} catch (JAXBException jaxbe) {
			size = -1;
		}
		
		return size;
	}
	
	protected void deleteLeastRecentlyUsed() {
		trace("FileStorageProvider.deleteLeastRecentlyUsed()");
		
		UUID targetId = null;
		long target = Long.MAX_VALUE;
		
		synchronized(activityRecord) {
			if (activityRecord.size() == 0) {
				throw new IllegalStateException("Filesystem and runtime are out of sync");
			}
			
			for (UUID id : activityRecord.keySet()) {
				if (target > activityRecord.get(id)) {
					targetId = id;
					target = activityRecord.get(id);
				}
			}
			
			if (targetId != null) {
				File[] files = new File[] {
						new File(getReservationFilePath(targetId)),
						new File(getMetadataFilePath(targetId)),
						new File(getDataFilePath(targetId))
				};
				
				for (File f : files) {
					if (f.exists()) {
						long size = f.length();
						
						if (f.delete()) {
							calculatedSize -= size;
						}
					}
				}
				
				activityRecord.remove(targetId);
			}
		}
	}
}